
def replace_arabic(string: str):
    return string.replace('ك', 'ک').replace('ي', 'ی').strip(" \u200c")


def replace_persian(string: str):
    return string.replace('ک', 'ك').replace('ی', 'ي').strip(" \u200c")



import asyncio
import datetime
import functools
import logging
import os
import re
from dataclasses import dataclass
from typing import List, Optional

import aiohttp
import bs4
import pandas as pd
import requests
from pytse_client import (
    config,
    symbols_data,
    translations,
    tse_settings,
    utils,
)
from pytse_client.download import download, download_ticker_client_types_record
from pytse_client.proxy.dto import ShareholderData
from pytse_client.proxy.tsetmc import get_day_shareholders_history
from pytse_client.ticker.api_extractors import (
    Order,
    TradeSummary,
    get_corporate_trade_summary,
    get_individual_trade_summary,
    get_orders,
)
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL
from pytse_client.utils.decorators import catch
from pytse_client.utils.persian import replace_persian, replace_arabic
from tenacity import retry, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)
logger.addHandler(logging.NullHandler())


@dataclass
class RealtimeTickerInfo:
    state: Optional[str]
    last_price: Optional[float]
    adj_close: Optional[float]
    yesterday_price: Optional[float]
    open_price: Optional[float]
    high_price: Optional[float]
    low_price: Optional[float]
    count: Optional[int]
    volume: Optional[int]
    value: Optional[int]
    last_date: Optional[datetime.datetime]
    best_demand_vol: Optional[int]
    best_demand_price: Optional[float]
    best_supply_vol: Optional[int]
    best_supply_price: Optional[float]
    sell_orders: Optional[List[Order]]
    buy_orders: Optional[List[Order]]
    individual_trade_summary: Optional[TradeSummary]
    corporate_trade_summary: Optional[TradeSummary]
    nav: Optional[int]
    nav_date: Optional[str]
    # ارزش بازار
    market_cap: Optional[int]


class Ticker:
    def __init__(
        self,
        symbol: str,
        index: Optional[str] = None,
        adjust: bool = False,
    ):
        self._index = index or symbols_data.get_ticker_index(symbol)
        self.symbol = symbol if index is None else self._index
        self.adjust = adjust
        self.daily_records_csv_path = (
            f"{config.DATA_BASE_PATH}/{self.symbol}.csv"
        )
        self.adjusted_daily_records_csv_path = (
            f"{config.DATA_BASE_PATH}/{self.symbol}-ت.csv"
        )
        self._url = tse_settings.TSE_TICKER_ADDRESS.format(self._index)
        self._info_url = tse_settings.TSE_ISNT_INFO_URL.format(self._index)
        self._client_types_url = TSE_CLIENT_TYPE_DATA_URL.format(self._index)
        # FIXME: if symbol equals instrument id, it cannot fetch introduction
        # page. it occurrs when Ticker called with index id not symbol
        self._introduction_url = (
            tse_settings.TSE_TICKER_INTRODUCTION_URL.format(
                replace_persian(self.symbol)
            )
        )
        self._history: pd.DataFrame = pd.DataFrame()

        if self.adjust:
            if os.path.exists(self.adjusted_daily_records_csv_path):
                self.from_file()
            else:
                self.from_web()
        else:
            if os.path.exists(self.daily_records_csv_path):
                self.from_file()
            else:
                self.from_web()

    def from_web(self):
        self._history = download(
            symbols=self.symbol,
            adjust=self.adjust,
        )[self.symbol]

    def from_file(self):
        if self.adjust:
            self._history = pd.read_csv(self.adjusted_daily_records_csv_path)
        else:
            self._history = pd.read_csv(self.daily_records_csv_path)
        self._history["date"] = pd.to_datetime(self._history["date"])

    @property
    def history(self):
        return self._history

    @property
    def url(self):
        return self._url

    @property
    def index(self):
        return self._index

    @property
    def instrument_id(self):
        """
        instrument id of a ticker is unique and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"InstrumentID='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def ci_sin(self):
        """
        instrument id of a ticker is like instrument_id and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"CIsin='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def title(self) -> str:
        return replace_arabic(
            re.findall(r"Title='(.*?)',", self._ticker_page_response.text
                       )[0].split("-")[0]
        )

    @property
    def fulltitle(self) -> str:
        return replace_arabic(
            re.findall(r"Title='(.*?)',", self._ticker_page_response.text)[0]
        )

    @property
    def group_name(self) -> str:
        return re.findall(
            r"LSecVal='([\D]*)',", self._ticker_page_response.text
        )[0]

    @property
    def p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have p/e
        """
        adj_close = self.get_ticker_real_time_info_response().adj_close
        eps = self.eps
        if adj_close is None or eps is None or eps == 0:
            return None
        return self.get_ticker_real_time_info_response().adj_close / self.eps

    @property
    def group_p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like وملت does not have group P/E (gpe)
        """
        gpe = re.findall(
            r"SectorPE='([\d.]*)',", self._ticker_page_response.text
        )
        if not gpe or not gpe[0]:
            return None
        return float(gpe[0])

    @property
    def eps(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have eps
        """
        eps = re.findall(
            r"EstimatedEPS='([-,\d]*)',", self._ticker_page_response.text
        )[0]
        if eps == "":
            return None
        return float(eps)

    @property
    def p_s_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have p/s
        """
        adj_close = self.get_ticker_real_time_info_response().adj_close
        psr = self.psr
        if adj_close is None or psr is None or psr == 0:
            return None
        return adj_close / psr

    @property
    def psr(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have psr
        """
        psr = re.findall(
            r"PSR='([+-]?[\.\d]*)'", self._ticker_page_response.text
        )
        if not psr or psr[0] == "":
            return None
        return float(psr[0])

    @property
    def total_shares(self) -> float:
        return float(
            re.findall(r"ZTitad=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def base_volume(self) -> float:
        return float(
            re.findall(r"BaseVol=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def fiscal_year(self) -> Optional[str]:
        fiscal_year = re.findall(
            r"سال مالی :<\/td>.*?>(.*?)<",
            self._ticker_introduction_page_response.text,
            re.DOTALL
        )
        return fiscal_year[0] if fiscal_year else None

    @property
    def flow(self) -> str:
        """
        عنوان بازار
        """
        flow_code = re.findall(
            r"Flow='(.*?)'", self._ticker_page_response.text
        )[0]
        return self._flow_name(flow_code)

    @property
    @catch(IndexError, ValueError)
    def sta_max(self) -> float:
        """
        حداکثر قیمت مجاز
        """
        return float(re.findall(
            r"PSGelStaMax='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    @catch(IndexError, ValueError)
    def sta_min(self) -> float:
        """
        حداقل قیمت مجاز
        """
        return float(re.findall(
            r"PSGelStaMin='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    @catch(IndexError, ValueError)
    def min_week(self) -> float:
        """
        حداقل قیمت هفته اخیر
        """
        return float(re.findall(
            r"MinWeek='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    @catch(IndexError, ValueError)
    def max_week(self) -> float:
        """
        حداکثر قیمت هفته اخیر
        """
        return float(re.findall(
            r"MaxWeek='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    @catch(IndexError, ValueError)
    def min_year(self) -> float:
        """
        حداقل قیمت بازه سال
        """
        return float(re.findall(
            r"MinYear='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    @catch(IndexError, ValueError)
    def max_year(self) -> float:
        """
        حداکثر قیمت بازه سال
        """
        return float(re.findall(
            r"MaxYear='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    def month_average_volume(self) -> str:
        """
        میانگین حجم ماه
        """
        return re.findall(
            r"QTotTran5JAvg='(.*?)'", self._ticker_page_response.text
        )[0]

    @property
    @catch(IndexError, ValueError)
    def float_shares(self) -> float:
        """
        درصد سهام شناور
        """
        return float(re.findall(
            r"KAjCapValCpsIdx='(.*?)'", self._ticker_page_response.text
        )[0])

    @property
    def client_types(self):
        return download_ticker_client_types_record(self._index)

    @property
    def trade_dates(self):
        return self._history["date"].to_list()

    @property
    def shareholders(self) -> pd.DataFrame:
        session = utils.requests_retry_session()
        page = session.get(self._shareholders_url, timeout=5)
        session.close()
        soup = bs4.BeautifulSoup(page.content, 'html.parser')
        table: bs4.PageElement = soup.find_all("table")[0]
        shareholders_df = utils.get_shareholders_html_table_as_csv(table)
        shareholders_df = shareholders_df.rename(
            columns=translations.SHAREHOLDERS_FIELD_MAPPINGS
        )
        return shareholders_df

    def get_shareholders_history(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None
    ) -> pd.DataFrame:
        """
            a helper function to use shareholders_history_async
        """
        return asyncio.run(
            self.get_shareholders_history_async(
                from_when,
                to_when,
                only_trade_days,
                session,
            ),
        )

    async def get_shareholders_history_async(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None,
    ) -> pd.DataFrame:
        requested_dates = utils.datetime_range(to_when - from_when, to_when)
        session_created = False
        if not session:
            session_created = True
            conn = aiohttp.TCPConnector(limit=25)
            session = aiohttp.ClientSession(connector=conn)
        tasks = []
        filtered_dates = list(
            filter(
                lambda date: not only_trade_days or
                (only_trade_days and date.date() in self.trade_dates),
                requested_dates,
            )
        )
        for date in filtered_dates:
            tasks.append(
                get_day_shareholders_history(self._index, date, session)
            )
        all_tickers_shareholders: List[List[ShareholderData]
                                       ] = await asyncio.gather(*tasks)
        if session_created is True:
            await session.close()

        rows = []
        for ticker_shareholders, info_date in zip(
            all_tickers_shareholders, filtered_dates
        ):
            for shareholder_data in ticker_shareholders:
                rows.append(
                    [
                        info_date,
                        shareholder_data.id,
                        shareholder_data.shares,
                        shareholder_data.percentage,
                        shareholder_data.instrument_id,
                        shareholder_data.name,
                        shareholder_data.change,
                    ]
                )
        return pd.DataFrame(
            data=rows,
            columns=[
                'date',
                'shareholder_id',
                'shareholder_shares',
                'shareholder_percentage',
                'shareholder_instrument_id',
                'shareholder_name',
                'change',
            ]
        )

    @property
    def state(self):
        return self.get_ticker_real_time_info_response().state

    @property
    def last_price(self):
        return self.get_ticker_real_time_info_response().last_price

    @property
    def adj_close(self):
        return self.get_ticker_real_time_info_response().adj_close

    @property
    def yesterday_price(self):
        return self.get_ticker_real_time_info_response().yesterday_price

    @property
    def open_price(self):
        return self.get_ticker_real_time_info_response().open_price

    @property
    def high_price(self):
        return self.get_ticker_real_time_info_response().high_price

    @property
    def low_price(self):
        return self.get_ticker_real_time_info_response().low_price

    @property
    def count(self):
        return self.get_ticker_real_time_info_response().count

    @property
    def volume(self):
        return self.get_ticker_real_time_info_response().volume

    @property
    def value(self):
        return self.get_ticker_real_time_info_response().value

    @property
    def last_date(self):
        return self.get_ticker_real_time_info_response().last_date

    @property
    def best_demand_vol(self):
        return self.get_ticker_real_time_info_response().best_demand_vol

    @property
    def best_demand_price(self):
        return self.get_ticker_real_time_info_response().best_demand_price

    @property
    def best_supply_vol(self):
        return self.get_ticker_real_time_info_response().best_supply_vol

    @property
    def best_supply_price(self):
        return self.get_ticker_real_time_info_response().best_supply_price

    @property
    def nav(self):
        return self.get_ticker_real_time_info_response().nav

    @property
    def market_cap(self):
        return self.get_ticker_real_time_info_response().market_cap

    @property
    def nav_date(self):
        return self.get_ticker_real_time_info_response().nav_date

    def get_ticker_real_time_info_response(self) -> RealtimeTickerInfo:
        """
        notes on usage:
        - Real time data might not be always available
        check for None values before usage
        """
        session = utils.requests_retry_session()
        response = session.get(self._info_url, timeout=5)
        session.close()

        response_sections_list = response.text.split(";")

        if len(response_sections_list) >= 1:
            price_section = response_sections_list[0].split(",")
            try:
                state = self._instrument_state(price_section[1])
                yesterday_price = int(price_section[5])
                open_price = int(price_section[4])
                high_price = int(price_section[6])
                low_price = int(price_section[7])
                count = int(price_section[8])
                volume = int(price_section[9])
                value = int(price_section[10])
                last_date = datetime.datetime.strptime(
                    price_section[12] + price_section[13],
                    '%Y%m%d%H%M%S'
                )
            except (ValueError, IndexError):
                state = None
                yesterday_price = None
                open_price = None
                high_price = None
                low_price = None
                count = None
                volume = None
                value = None
                last_date = None

            # in some cases last price or adj price is undefined
            try:
                last_price = int(price_section[2])
            # when instead of number value is `F`
            except (ValueError, IndexError):
                last_price = None
            try:
                adj_close = int(price_section[3])
            except (ValueError, IndexError):
                adj_close = None
            try:
                market_cap = adj_close * self.total_shares
            except ValueError:
                market_cap = None

        try:
            info_section = response_sections_list[0].split(",")
            nav = int(info_section[15])
            nav_date = str(info_section[14])
        except (ValueError, IndexError):
            nav = None
            nav_date = None

        try:
            orders_section = response_sections_list[2]
            buy_orders, sell_orders = get_orders(orders_section)
            best_demand_vol = (
                buy_orders[0].volume if 0 < len(buy_orders) else None
            )
            best_demand_price = (
                buy_orders[0].price if 0 < len(buy_orders) else None
            )
            best_supply_vol = (
                sell_orders[0].volume if 0 < len(sell_orders) else None
            )
            best_supply_price = (
                sell_orders[0].price if 0 < len(sell_orders) else None
            )
        except (IndexError):
            buy_orders = []
            sell_orders = []
            logger.warning(
                f"""not enough sections in response to demand
                and supply information resp: {response_sections_list}"""
            )
            best_demand_vol = None
            best_demand_price = None
            best_supply_vol = None
            best_supply_price = None
            buy_orders = None
            sell_orders = None

        if len(response_sections_list) >= 4:
            trade_summary_section = response_sections_list[4]
            individual_trade_summary = get_individual_trade_summary(
                trade_summary_section
            )
            corporate_trade_summary = get_corporate_trade_summary(
                trade_summary_section
            )
        else:
            logger.warning(
                f"""not enough sections in response to extract trade summaries
                resp: {response_sections_list}"""
            )
            individual_trade_summary = None
            corporate_trade_summary = None

        return RealtimeTickerInfo(
            state,
            last_price,
            adj_close,
            yesterday_price,
            open_price,
            high_price,
            low_price,
            count,
            volume,
            value,
            last_date,
            best_demand_vol=best_demand_vol,
            best_demand_price=best_demand_price,
            best_supply_vol=best_supply_vol,
            best_supply_price=best_supply_price,
            buy_orders=buy_orders,
            sell_orders=sell_orders,
            individual_trade_summary=individual_trade_summary,
            corporate_trade_summary=corporate_trade_summary,
            nav=nav,
            market_cap=market_cap,
            nav_date=nav_date,
        )

    @property
    @functools.lru_cache()
    def _ticker_page_response(self):
        return utils.requests_retry_session().get(self._url, timeout=10)

    @functools.cached_property
    def _ticker_introduction_page_response(self):
        return utils.requests_retry_session().get(
            self._introduction_url, timeout=10
        )

    @functools.lru_cache()
    @retry(
        wait=wait_random(min=3, max=5),
        before_sleep=before_sleep_log(logger, logging.ERROR)
    )
    async def _get_ticker_daily_info_page_response(
        self, session, date
    ) -> requests.Response:
        async with session.get(
            tse_settings.INSTRUMENT_DAY_INFO_URL.format(
                index=self.index, date=date
            ),
        ) as response:
            response.raise_for_status()
            page = await response.text()
            logger.info(f"fetched date {date}")
            return page

    @property
    def _shareholders_url(self) -> str:
        return tse_settings.TSE_SHAREHOLDERS_URL.format(self.ci_sin)

    def _instrument_state(self, state_code) -> str:
        states = {
            "I ": "ممنوع",
            "A ": "مجاز",
            "AG": "مجاز-مسدود",
            "AS": "مجاز-متوقف",
            "AR": "مجاز-محفوظ",
            "IG": "ممنوع-مسدود",
            "IS": "ممنوع-متوقف",
            "IR": "ممنوع-محفوظ",
        }
        return states.get(state_code, "")

    def _flow_name(self, flow_code) -> str:
        flows = {
            "0": "عمومی - مشترک بین بورس و فرابورس",
            "1": "بورس",
            "2": "فرابورس",
            "3": "آتی",
            "4": "پایه فرابورس",
            "5": "پایه فرابورس (منتشر نمی شود)",
        }
        return flows.get(flow_code, "")

    def get_trade_details(self):
        session = utils.requests_retry_session()
        page = session.get(tse_settings.TSE_TRADE_DETAIL_URL.format(
            self.index), timeout=5)
        session.close()
        soup = bs4.BeautifulSoup(page.content, "lxml")
        xml_rows = soup.find_all("row")
        rows = []
        for xml_row in xml_rows:
            cells = xml_row.find_all("cell")
            row = [
                datetime.time.fromisoformat(cells[1].text),
                int(cells[2].text),
                float(cells[3].text)
            ]
            rows.append(row)
        return pd.DataFrame(rows, columns=['date', 'volume', 'price'])



import datetime
from pytse_client import Ticker

symbols = [
    "شکبیر",
    "برکت",
]
from_times = [
    datetime.timedelta(days=90),
    datetime.timedelta(days=150),
]
for symbol, from_when in zip(symbols, from_times):
    print(f"downloading {symbol}")
    Ticker(symbol).get_shareholders_history(from_when=from_when
                                            ).to_csv(f"{symbol}.csv")
    print(f"downloaded {symbol}")



import logging
from concurrent import futures
from io import StringIO
from pathlib import Path
from typing import Dict, List, Optional, Union

import re
import jdatetime
import pandas as pd
from pytse_client import config, symbols_data, translations, tse_settings
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL
from pytse_client.utils import persian, requests_retry_session
from pytse_client.scraper.symbol_scraper import MarketSymbol
from requests import HTTPError
from requests.sessions import Session
from tenacity import retry, retry_if_exception_type, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)


def _handle_ticker_index(symbol):
    ticker_index = symbols_data.get_ticker_index(symbol)

    if ticker_index is None:
        market_symbol = get_symbol_info(symbol)
        if market_symbol is not None:
            symbols_data.append_symbol_to_file(market_symbol)
            ticker_index = market_symbol.index
    return ticker_index


def _extract_ticker_client_types_data(ticker_index: str) -> List:
    url = TSE_CLIENT_TYPE_DATA_URL.format(ticker_index)
    with requests_retry_session() as session:
        response = session.get(url, timeout=5)
    data = response.text.split(";")
    data = [row.split(",") for row in data]
    return data


def _create_financial_index_from_text_response(data):
    data = re.split(r'\;|\,', data)
    dates = data[::2]
    values = data[1::2]
    values = list(map(float, values))
    df = pd.DataFrame(
        tuple(zip(dates, values)), columns=['jdate', 'value'])
    return df


def _adjust_data_frame(df, include_jdate):
    df.date = pd.to_datetime(df.date, format="%Y%m%d")
    if include_jdate:
        df['jdate'] = ""
        df.jdate = df.date.apply(
            lambda gregorian: jdatetime.date.fromgregorian(date=gregorian)
        )


def _adjust_data_frame_for_fIndex(df, include_jdate):
    df["date"] = df["jdate"].apply(
        lambda x: jdatetime.datetime.
        togregorian(jdatetime.datetime.strptime(x, "%Y/%m/%d")))
    df["date"] = pd.to_datetime(df["date"], format="%Y%m%d")
    if include_jdate:
        df['jdate'] = ""
        df.jdate = df.date.apply(
            lambda gregorian: jdatetime.date.fromgregorian(date=gregorian)
        )
    else:
        df.drop(columns=["jdate"], inplace=True)


def download(
    symbols: Union[List, str],
    write_to_csv: bool = False,
    include_jdate: bool = False,
    base_path: str = config.DATA_BASE_PATH,
    adjust: bool = False,
) -> Dict[str, pd.DataFrame]:
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        session = requests_retry_session()
        for symbol in symbols:
            if (symbol.isnumeric()
                    and symbols_data.get_ticker_index(symbol) is None):
                ticker_indexes = [symbol]
            else:
                ticker_index = _handle_ticker_index(symbol)
                if ticker_index is None:
                    raise Exception(f"Cannot find symbol: {symbol}")
                ticker_indexes = symbols_data.get_ticker_old_index(
                    symbol)
                ticker_indexes.insert(0, ticker_index)

            for index in ticker_indexes:
                future = executor.submit(
                    download_ticker_daily_record, index, session
                )

                future_to_symbol[future] = symbol

        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            try:
                df: pd.DataFrame = future.result()
            except pd.errors.EmptyDataError as ex:
                logger.error(
                    f"Cannot read daily trade records for symbol: {symbol}",
                    extra={"Error": ex}
                )
                continue
            df = df.iloc[::-1].reset_index(drop=True)
            df = df.rename(columns=translations.HISTORY_FIELD_MAPPINGS)
            df = df.drop(columns=["<PER>", "<TICKER>"])
            _adjust_data_frame(df, include_jdate)

            if symbol in df_list:
                df_list[symbol] = (
                    df_list[symbol].append(
                        df,
                        ignore_index=True,
                        sort=False
                    ).sort_values('date').reset_index(drop=True)
                )
            else:
                df_list[symbol] = df

            if adjust:
                df_list[symbol] = adjust_price(df_list[symbol])

            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                if adjust:
                    df_list[symbol].to_csv(
                        f'{base_path}/{symbol}-ت.csv',
                        index=False
                    )
                else:
                    df_list[symbol].to_csv(
                        f'{base_path}/{symbol}.csv',
                        index=False
                    )

    if len(df_list) != len(symbols):
        print("Warning, download did not complete, re-run the code")
    session.close()
    return df_list


def adjust_price(
    df: pd.DataFrame
) -> pd.DataFrame:
    """
    Adjust historical records of stock

    There is a capital increase/profit sharing,
    if today "Final Close Price" is not equal to next day
    "Yesterday Final Close Price" by using this ratio,
    performance adjustment of stocks is achieved

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with historical records.

    Returns
    -------
    pd.DataFrame
        DataFrame with adjusted historical records.

    Notes
    -----
    DataFrame can not be empty or else it makes runtime error
    Type of DataFrame must be RangeIndex to make proper range of records
    that need to be modified

    diff: list
        list of indexs of the day after capital increase/profit sharing
    ratio_list: List
        List of ratios to adjust historical data of stock
    ratio: Float
        ratio = df.loc[i].adjClose / df.loc[i+1].yesterday

    Description
    -----------
    # Note: adjustment does not include Tenth and twentieth days
    df.index = range(0,101,1)
    # step is 1
    step = df.index.step
    diff = [10,20]
    ratio_list = [0.5, 0.8]
    df.loc[0:10-step, [open,...]] * ratio[0]
    df.loc[10:20-step, [open,...]] * ratio[1]
    """
    if df.empty or not isinstance(df.index, pd.core.indexes.range.RangeIndex):
        return df

    new_df = df.copy()
    step = new_df.index.step
    diff = list(new_df.index[new_df.shift(1).adjClose != new_df.yesterday])
    if len(diff) > 0:
        diff.pop(0)
    ratio = 1
    ratio_list = []
    for i in diff[::-1]:
        ratio *= (
            new_df.loc[i, 'yesterday'] / new_df.shift(1).loc[i, 'adjClose']
        )
        ratio_list.insert(0, ratio)
    for i, k in enumerate(diff):
        if i == 0:
            start = new_df.index.start
        else:
            start = diff[i-1]
        end = diff[i]-step
        new_df.loc[start:end, [
            'open',
            'high',
            'low',
            'close',
            'adjClose',
            'yesterday',
        ]] = round(
            new_df.loc[start:end, [
                'open',
                'high',
                'low',
                'close',
                'adjClose',
                'yesterday',
            ]] * ratio_list[i]
        )

    return new_df


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_random(min=1, max=4),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def download_ticker_daily_record(ticker_index: str, session: Session):
    url = tse_settings.TSE_TICKER_EXPORT_DATA_ADDRESS.format(ticker_index)
    response = session.get(url, timeout=10)
    if 400 <= response.status_code < 500:
        logger.error(
            f"Cannot read daily trade records from the url: {url}",
            extra={
                "response": response.text,
                "code": response.status_code
            }
        )

    response.raise_for_status()

    data = StringIO(response.text)
    return pd.read_csv(data)


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_random(min=1, max=4),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def download_fIndex_record(fIndex: str, session: Session):
    url = tse_settings.TSE_FINANCIAL_INDEX_EXPORT_DATA_ADDRESS.format(fIndex)
    response = session.get(url, timeout=10)
    if 400 <= response.status_code < 500:
        logger.error(
            f"Cannot read daily trade records from the url: {url}",
            extra={
                "response": response.text,
                "code": response.status_code
            }
        )

    response.raise_for_status()
    data = response.text
    if not data or ";" not in data or "," not in data:
        raise ValueError(f"""Invalid response from the url: {url}.
                         \nExpected valid financial index data.""")
    df = _create_financial_index_from_text_response(data)
    return df


def download_financial_indexes(
    symbols: Union[List, str],
    write_to_csv: bool = False,
    include_jdate: bool = False,
    base_path: str = config.FINANCIAL_INDEX_BASE_PATH,
) -> Dict[str, pd.DataFrame]:
    if symbols == "all":
        symbols = symbols_data.all_financial_index()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        session = requests_retry_session()
        for symbol in symbols:
            if (
                symbol.isnumeric()
                and
                symbols_data.get_financial_index(symbol) is None
            ):
                financial_index = symbol
            else:
                financial_index = symbols_data.get_financial_index(symbol)
            if financial_index is None:
                raise Exception(f"Cannot find financial index: {symbol}")

            future = executor.submit(
                download_fIndex_record, financial_index, session
            )

            future_to_symbol[future] = symbol

        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            try:
                df: pd.DataFrame = future.result()
            except pd.errors.EmptyDataError as ex:
                logger.error(
                    f"Cannot read daily trade records for symbol: {symbol}",
                    extra={"Error": ex}
                )
                continue
            _adjust_data_frame_for_fIndex(df, include_jdate)
            df_list[symbol] = df

            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                df_list[symbol].to_csv(
                    f'{base_path}/{symbol}.csv',
                    index=False
                )

    if len(df_list) != len(symbols):
        print("Warning, download did not complete, re-run the code")
    session.close()
    return df_list


def download_client_types_records(
    symbols: Union[List, str],
    write_to_csv: bool = False,
    include_jdate: bool = False,
    base_path: str = config.CLIENT_TYPES_DATA_BASE_PATH
):
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        for symbol in symbols:
            ticker_index = _handle_ticker_index(symbol)
            if ticker_index is None:
                raise Exception(f"Cannot find symbol: {symbol}")
            future = executor.submit(
                download_ticker_client_types_record, ticker_index
            )
            future_to_symbol[future] = symbol
        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            df: pd.DataFrame = future.result()
            # ignore failures
            if df is None:
                continue
            _adjust_data_frame(df, include_jdate)
            df_list[symbol] = df
            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                df.to_csv(f'{base_path}/{symbol}.csv')

    if len(df_list) != len(symbols):
        print(
            """could not download client types for all the symbols make
        sure you have what you need or re-run the function"""
        )
    return df_list


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_random(min=1, max=4),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def download_ticker_client_types_record(ticker_index: Optional[str]):
    data = _extract_ticker_client_types_data(ticker_index)
    if len(data) == 1:
        logger.warning(
            f"""Cannot create client types data for ticker{ticker_index}
             from data: {data}""",
            extra={"ticker_index": ticker_index}
        )
        return None
    client_types_data_frame = pd.DataFrame(
        data,
        columns=[
            "date", "individual_buy_count", "corporate_buy_count",
            "individual_sell_count", "corporate_sell_count",
            "individual_buy_vol", "corporate_buy_vol", "individual_sell_vol",
            "corporate_sell_vol", "individual_buy_value",
            "corporate_buy_value", "individual_sell_value",
            "corporate_sell_value"
        ]
    )
    for i in [
        "individual_buy_", "individual_sell_", "corporate_buy_",
        "corporate_sell_"
    ]:
        client_types_data_frame[f"{i}mean_price"] = (
            client_types_data_frame[f"{i}value"].astype(float) /
            client_types_data_frame[f"{i}vol"].astype(float)
        )
    client_types_data_frame["individual_ownership_change"] = (
        client_types_data_frame["corporate_sell_vol"].astype(float) -
        client_types_data_frame["corporate_buy_vol"].astype(float)
    )
    return client_types_data_frame


def get_symbol_id(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception("Sorry, tse server did not respond")

    symbol_full_info = response.text.split(';')[0].split(',')
    if persian.replace_arabic(symbol_name) == symbol_full_info[0].strip():
        return symbol_full_info[2]  # symbol id
    return None


def get_symbol_info(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception(f"{symbol_name}: Sorry, tse server did not respond")

    symbols = response.text.split(';')
    market_symbol = MarketSymbol(
        code=None,
        symbol=None,
        index=None,
        name=None,
        old=[],
    )
    for symbol_full_info in symbols:
        if symbol_full_info.strip() == "":
            continue
        symbol_full_info = symbol_full_info.split(',')
        if persian.replace_arabic(symbol_full_info[0]) == symbol_name:
            # if symbol id is active
            if symbol_full_info[7] == '1':
                market_symbol.symbol = persian.replace_arabic(
                    symbol_full_info[0]
                )
                market_symbol.name = persian.replace_arabic(
                    symbol_full_info[1]
                )
                market_symbol.index = symbol_full_info[2]  # active symbol id
            else:
                market_symbol.old.append(symbol_full_info[2])  # old symbol id

    if market_symbol.index is None:
        return None
    return market_symbol



import re
from typing import List, Optional


def scrape_daily_info_page_for_date(
    daily_info_page: str
) -> Optional[str]:
    daily_info_page_date = re.findall("DEven='(.*?)'", daily_info_page)
    if len(daily_info_page_date) != 1:
        return
    return daily_info_page_date[0]



# from .requests import requests_retry_session
from .string import convert_to_number_if_number
from typing import List, Tuple
import bs4
import pandas as pd


def get_html_table_header_and_rows(
    table: bs4.PageElement
) -> Tuple[List, List]:
    """
    return header and rows from a html table as a list
    """
    header = []
    rows = []
    table_header = table.find("tr")
    table_rows = table.find_all("tr")[1:]
    for items in table_header:
        header.append(items.get_text())

    for table_row in table_rows:
        row = []
        for cell in table_row.findAll(['th', 'td']):
            row.append(cell)
        rows.append(row)

    return header, rows


def get_shareholders_html_table_as_csv(table) -> pd.DataFrame:
    """
    given table element from shareholders page returns DatFrame
    Containing the table
    """
    header, rows = get_html_table_header_and_rows(table)
    df_rows = []

    for row in rows:
        df_row = []
        for cell in row:
            cell_div = cell.find("div")
            if cell_div and cell_div.get_text() != "":
                df_row.append(convert_to_number_if_number(cell_div["title"]))
            else:
                df_row.append(
                    convert_to_number_if_number(cell.get_text().strip())
                )
        df_rows.append(df_row)
    return pd.DataFrame(data=df_rows, columns=header)



from pytse_client import FinancialIndex

fIndex1 = FinancialIndex("شاخص كل")
# شاخص کل
fIndex2 = FinancialIndex(symbol="",
                         index="32097828799138957")
# فرآوده های نفتی
fIndex3 = FinancialIndex(symbol="",
                         index="12331083953323969")
fIndex4 = FinancialIndex(symbol="شاخص قيمت 50 شركت")
fIndex5 = FinancialIndex(symbol="فني مهندسي")

fIndex = FinancialIndex(symbol="فراورده نفتي",
                        index="12331083953323969")
# اگر باهم استفاده شوند معیار index میشود.
print(fIndex.history)
print(fIndex.intraday_price)
print(fIndex.high)
print(fIndex.low)
print(fIndex.last_update)
print(fIndex.last_value)
print(fIndex.contributing_symbols)



"""
    This example shows how to download and get tse tickers
    client types records specifically (as a 'Pandas DataFrame' object),
    and easily save them in .csv format.
"""
from pytse_client import download_client_types_records

if __name__ == '__main__':
    # Download all tickers clients types at once.
    # This function returns a dictionary.
    records_1 = download_client_types_records("all")

    # Download specific tickers client types records
    records_2 = download_client_types_records(["فولاد", "ذوب"])
    print(records_2["فولاد"])
    print("Fuldad: ", records_2["فولاد"], sep="\n")
    print("Zoob: ", records_2["ذوب"], sep="\n")

    # Download and save tickers client types records
    records_3 = download_client_types_records("فولاد", write_to_csv=True)
    print("Foolad:", records_3["فولاد"], sep="\n")



import requests as req
import re
import json


if __name__ == '__main__':
    text = req.get(
        "http://www.tsetmc.com/Loader.aspx?Partree=151315&Flow=1").text

    market_indexes = re.findall(
        r'<tr><td><a target="_blank" href="Loader'
        r'\.aspx\?ParTree=15131J&i=\d*">.*</a></td>',
        text)
    final_dict = {}

    for market_index in market_indexes:
        key = re.search(
            '">(.*)</a></td>', market_index).group(1)
        index = re.search(
            '<tr><td><a target="_blank" href="Loader'
            r'\.aspx\?ParTree=15131J&i=(\d*)',
            market_index).group(1)
        section_number = re.search(r'(^\d+|\d+$)', key)
        section_number = int(section_number.group(
            1)) if section_number is not None else ""
        key = re.sub(r'^\d+|\d+$|-', '', key)
        final_dict[key] = {"index": index, "section_number": section_number}

    with open("indices_name.json", "w") as f:
        json.dump(final_dict, f, indent=2, ensure_ascii=False)



import functools
import pandas as pd

from typing import Optional, List, Dict
from bs4 import BeautifulSoup

from pytse_client import utils
from pytse_client import symbols_data
from pytse_client import tse_settings
from pytse_client.download import download_financial_indexes
from pytse_client import config


class FinancialIndex:
    def __init__(self,
                 symbol: str,
                 index: Optional[str] = None,
                 base_path: Optional[str] = config.FINANCIAL_INDEX_BASE_PATH,
                 write_history: Optional[bool] = False,
                 ):
        self._index: str = index or symbols_data.get_financial_index(symbol)
        self.symbol: str = symbol if index is None else self._index
        self._intraday_url: str = tse_settings\
            .TSE_FINANCIAL_INDEX_EXPORT_INTRADAY_DATA_ADDRESS\
            .format(self._index)
        self._base_path = base_path
        self._write_history = write_history

    @property
    def last_update(self):
        # زمان انتشار
        soup = self._financial_index_page_soup
        before_update_time_td: BeautifulSoup = \
            soup.find_all("td", text="زمان انتشار")[0]
        update_time_td: BeautifulSoup =\
            before_update_time_td.find_next_siblings("td")[0]
        return pd.to_datetime(update_time_td.get_text()).strftime("%H:%M")

    @property
    def last_value(self):
        # آخرین مقدار شاخص
        soup = self._financial_index_page_soup
        before_lastval_td: BeautifulSoup =\
            soup.find_all("td", text="آخرین مقدار شاخص")[0]
        lastval_td: BeautifulSoup =\
            before_lastval_td.find_next_siblings("td")[0]
        return float(lastval_td.get_text().replace(",", ""))

    @property
    def high(self):
        # بیشترین مقدار روز
        soup = self._financial_index_page_soup
        before_high_td: BeautifulSoup =\
            soup.find_all("td", text="بیشترین مقدار روز")[0]
        high_td: BeautifulSoup =\
            before_high_td.find_next_siblings("td")[0]
        return high_td.get_text()

    @property
    def low(self):
        # کمترین مقدار روز
        soup = self._financial_index_page_soup
        before_low_td: BeautifulSoup =\
            soup.find_all("td", text="کمترین مقدار روز")[0]
        low_td: BeautifulSoup =\
            before_low_td.find_next_siblings("td")[0]
        return low_td.get_text()

    @property
    @functools.lru_cache()
    def history(self):
        return download_financial_indexes(self.symbol,
                                          write_to_csv=self._write_history,
                                          include_jdate=True,
                                          base_path=self._base_path
                                          )[self.symbol]

    def _get_contributing_symbols(self):
        # شرکت های موجود در شاخص
        soup = self._financial_index_page_soup
        before_contr_symbols: BeautifulSoup =\
            soup.find_all("div", text="شرکت های موجود در شاخص")[0]
        contr_symbols: BeautifulSoup =\
            before_contr_symbols.find_next_siblings("div")[0]
        _index_symbols = list(map(lambda x: x.a["href"].split("i=")[
                              1], contr_symbols.find_all("td")[::9]))
        _symbols = list(map(lambda x: x.get_text(),
                        contr_symbols.find_all("td")[::9]))
        _contr_symbols = list(zip(_index_symbols, _symbols))
        return [
            {
                "symbol": _contr_symbol[1],
                "index": _contr_symbol[0]
            } for _contr_symbol in _contr_symbols
        ]

    @property
    @functools.lru_cache()
    def contributing_symbols(self) -> List[Dict[str, str]]:
        return self._get_contributing_symbols()

    @property
    def intraday_price(self) -> pd.DataFrame:
        soup = self._financial_index_page_soup
        before_intraday_price: BeautifulSoup =\
            soup.find_all("div", text="سابقه شاخص روز جاری")[0]

        intraday_price: BeautifulSoup =\
            before_intraday_price.find_next_siblings("div")[0]
        intraday_price_ls = list(
            map(lambda bs: bs.get_text(), intraday_price.find_all("td"))
        )
        columns = ["time", "value", "change_percentage", "low", "high"]

        rows = self._get_rows(intraday_price_ls, len(columns))
        df = pd.DataFrame(rows, columns=columns)
        df = df.astype(
            {
                "value": float,
                "change_percentage": float,
                "low": float,
                "high": float
            },
            errors='raise'
        )

        df["time"] = pd.to_datetime(df["time"], format="%H:%M")\
            .dt.time
        df.set_index("time", inplace=True)
        return df

    @property
    def _financial_index_page_soup(self):
        return BeautifulSoup(self._financial_index_page_text, 'html.parser')

    @property
    def _financial_index_page_text(self):
        raw_html = utils.requests_retry_session()\
            .get(self._intraday_url, timeout=10).text
        return raw_html

    def _get_rows(self, intraday_price_ls, col_len):
        '''
        intraday_price_ls: a table in financial index history page that contains 
            intraday information as string
        col_len: number of table columns
        
        in the result return from tsetmc negative values like -3
        are represented as (3) here they are converted to normal representaion
        '''
        
        # Remove separators in numbers: 3,000 -> 3000
        intraday_price_ls = list(
            map(lambda x: x.replace(",", ""), intraday_price_ls))
        intraday_price_ls = list(
            map(lambda x: x.replace("(", "-"), intraday_price_ls))
        intraday_price_ls = list(
            map(lambda x: x.replace(")", ""), intraday_price_ls))
        rows = [intraday_price_ls[i:i+col_len]
                for i in range(0, len(intraday_price_ls), col_len)]
        return rows



from key_stats import filter_key_value
import requests
import re
try:
    response = requests.get(
        'http://redirectcdn.tsetmc.com'
        '/Site.aspx?ParTree=151713')
except Exception as e:
    print('--- Could not retrieve key stats documentation ---')
    raise e
text = response.text
key_stats_ls = [line for line in text.split(
    '\n') if re.search(r'is[0-9]+', line)]
nums_and_disc = []
for line in key_stats_ls:
    number, disc = re.split(r'\]|\[|\&', line)[1:3]
    number = int(number[2:])
    disc = re.split(r"\<\/div\>", disc)[0]
    nums_and_disc.append((number, disc))
with open("README.md", "w") as f2:
    f2.write(
        "در این بخش آمار کلیدی را در قالب"
        " یک نگاشت سه عضوی مشاهده می‌کنید."
        )
    f2.write("\n\n\n| شماره | توضیحات | سمبل |\n")
    f2.write("| --- | --- | --- |\n")
    for number, persian_explanation in nums_and_disc:
        symbol = filter_key_value[number]
        f2.write(f"| {number} | {persian_explanation} | {symbol} |\n")



"""
This example is about getting data for a ticker.
use this example if you want to get data for one ticker
"""

from pytse_client import Ticker, download
import pandas as pd

# to be able to see whole DataFrame columns
pd.set_option('display.max_columns', 20)

download(symbols="نوری", write_to_csv=True)  # optional
ticker = Ticker("نوری")
print(ticker.history)  # سابقه قیمت سهم
print(ticker.client_types)  # حقیقی حقوقی
print(ticker.title)  # نام شرکت
print(ticker.url)  # آدرس صفحه سهم
print(ticker.group_name)  # نام گروه
print(ticker.fiscal_year)  # سال مالی
print(ticker.eps)  # eps
print(ticker.p_e_ratio)  # P/E
print(ticker.group_p_e_ratio)  # group P/E
print(ticker.nav)  # NAV خالص ارزش دارایی‌ها ویژه صندوق‌ها می‌باشد
print(ticker.nav_date)  # last date of NAV ویژه صندوق‌ها می‌باشد
print(ticker.psr)  # PSR این نسبت ویژه شرکت‌های تولیدی است
print(ticker.p_s_ratio)  # P/S این نسبت ویژه شرکت‌های تولیدی است
print(ticker.base_volume)  # حجم مبنا
print(ticker.state)  # وضعیت نماد
print(ticker.last_price)  # قیمت آخرین معامله
print(ticker.adj_close)  # قیمت پایانی
print(ticker.market_cap)  # ارزش بازار شرکت
print(ticker.yesterday_price)  # قیمت دیروز
print(ticker.open_price)  # قیمت اولین معامله
print(ticker.high_price)  # قیمت حداکثر
print(ticker.low_price)  # قیمت حداقل
print(ticker.count)  # تعداد معاملات
print(ticker.volume)  # حجم معاملات
print(ticker.value)  # ارزش معاملات
print(ticker.last_date)  # آخرین اطلاعات قیمت ناشی از تغییرات شرکتی و معاملات
print(ticker.flow)  # عنوان بازار
print(ticker.sta_max)  # حداکثر قیمت مجاز
print(ticker.sta_min)  # حداقل قیمت مجاز
print(ticker.min_week)  # حداقل قیمت هفته اخیر
print(ticker.max_week)  # حداکثر قیمت هفته اخیر
print(ticker.min_year)  # حداقل قیمت بازه سال
print(ticker.max_year)  # حداکثر قیمت بازه سال
print(ticker.month_average_volume)  # میانگین حجم ماه
print(ticker.float_shares)  # درصد سهام شناور
print(ticker.total_shares)  # تعداد سهام
print(ticker.shareholders)  # اطلاعات سهام داران عمده
print(ticker.shareholders.percentage.sum())  # جمع سهام داران
print(ticker.get_trade_details())  # ریز معاملات روز جاری

real_time_data = ticker.get_ticker_real_time_info_response()
print(real_time_data.individual_trade_summary.buy_count)
print(real_time_data.individual_trade_summary.buy_vol)
print(real_time_data.individual_trade_summary.sell_count)
print(real_time_data.individual_trade_summary.sell_vol)
print(real_time_data.corporate_trade_summary.buy_count)
print(real_time_data.corporate_trade_summary.buy_vol)
print(real_time_data.corporate_trade_summary.sell_count)
print(real_time_data.corporate_trade_summary.sell_vol)



import json
from pathlib import Path
import os
import re
import pandas as pd
from typing import Dict, List, Tuple
from pytse_client import (
    utils,
    tse_settings,
    config
)

from pytse_client.ticker_statisticals.utils import get_index_to_symbol_map
from pytse_client.ticker_statisticals import (
    filter_key_value,
    filter_value_NONE
)


def _map_index_to_symbols():
    parent_dir = os.path.dirname(os.path.abspath(__file__))
    symbols_file = os.path.join(parent_dir, "data/symbols_name.json")
    with open(symbols_file, 'r') as json_file:
        symbol_dic = json.load(json_file)
    return get_index_to_symbol_map(symbol_dic)


def _get_list_of_processed_stats(raw_key_stats: str) \
        -> Tuple[List[str], List[str]]:

    # group 1 is idx, group 2 is key, group 3 is value
    proccessed_key_stats = re.sub(
        r'([0-9]+)\,([0-9]+)\,([0-9\.]+)\;',
        '@\\g<1>@\\g<2>,\\g<3>;', raw_key_stats)
    list_of_key_stats = proccessed_key_stats.split("@")[1:]
    indices = list_of_key_stats[0::2]
    values = list_of_key_stats[1::2]

    return indices, values


def get_aggregated_key_stats(base_path=None, to_csv=False)\
        -> Dict[str, Dict[str, str]]:
    aggregated_key_stats = {}
    index_to_symbol_map = _map_index_to_symbols()
    session = utils.requests_retry_session()
    try:
        response = session.get(tse_settings.KEY_STATS_URL)
    except Exception as e:
        raise Exception(f"Failed to get key stats: {e}")
    finally:
        session.close()

    raw_key_stats = response.text
    indices, values = _get_list_of_processed_stats(raw_key_stats)

    linked_stats = zip(indices, values)
    for idx_stat, val_stat in linked_stats:
        if idx_stat not in index_to_symbol_map:
            continue
        else:
            symbol = index_to_symbol_map[idx_stat]["symbol"]
            name = index_to_symbol_map[idx_stat]["name"]
        filter_key_found = {}
        segmented_val_stat = re.split(r'\;', val_stat)
        segmented_val_stat = list(
            filter(lambda x: x != '', segmented_val_stat))
        for each_segment in segmented_val_stat:
            key, val = each_segment.split(",")
            filter_key_found[filter_key_value[int(key)]] = val
        aggregated_key_stats[idx_stat] = {
            **filter_value_NONE,
            **filter_key_found,
            "symbol": symbol,
            "name": name
        }

    aggregated_key_stats_df = pd.DataFrame.from_dict(aggregated_key_stats).T

    # change type of columns to int
    str_cols = {"symbol", "name"}
    numeric_cols = set(aggregated_key_stats_df.columns) - str_cols
    numeric_cols_ls = list(numeric_cols)
    aggregated_key_stats_df[numeric_cols_ls]\
        = aggregated_key_stats_df[numeric_cols_ls].apply(
            pd.to_numeric,
            errors='coerce'
    )
    aggregated_key_stats_df["index"] = aggregated_key_stats_df.index
    aggregated_key_stats_df.reset_index(drop=True, inplace=True)

    if to_csv:
        base_path = config.KEY_STATS_BASE_PATH if\
            base_path is None else base_path
        Path(base_path).mkdir(parents=True, exist_ok=True)

        path = os.path.join(base_path, "key_stats.csv")
        aggregated_key_stats_df.to_csv(path, index=False)

    return aggregated_key_stats_df



import json
from typing import Dict, Set

from pytse_client import config
from pytse_client.scraper.symbol_scraper import MarketSymbol

ticker_name_to_index_mapping = None
financial_index_name_to_index_mapping = None


def financial_indexes_information() -> Dict[str, Dict]:
    global financial_index_name_to_index_mapping
    if financial_index_name_to_index_mapping is None:
        with open(
            f"{config.pytse_dir}/data/indices_name.json", "r", encoding="utf8"
        ) as symbols_name:
            financial_index_name_to_index_mapping = json.load(symbols_name)
    return financial_index_name_to_index_mapping


def symbols_information() -> Dict[str, Dict]:
    global ticker_name_to_index_mapping
    if ticker_name_to_index_mapping is None:
        with open(
            f"{config.pytse_dir}/data/symbols_name.json", "r", encoding="utf8"
        ) as symbols_name:
            ticker_name_to_index_mapping = json.load(symbols_name)
    return ticker_name_to_index_mapping


def get_financial_index(financial_index_name: str):
    return financial_indexes_information().get(
        financial_index_name, {}).get("index")


def get_ticker_index(ticker_symbol: str):
    return symbols_information().get(ticker_symbol, {}).get("index")


def get_ticker_old_index(ticker_symbol: str):
    return symbols_information().get(ticker_symbol, {}).get("old", []).copy()


def all_symbols() -> Set:
    return set(symbols_information().keys())


def all_financial_index() -> Set:
    return set(financial_indexes_information().keys())


def append_symbol_to_file(
    market_symbol: MarketSymbol,
):
    global ticker_name_to_index_mapping
    new_symbol = {
        market_symbol.symbol: {
            "index": market_symbol.index,
            "code": market_symbol.code,
            "name": market_symbol.name,
            "old": market_symbol.old
        }
    }
    if ticker_name_to_index_mapping is not None:
        ticker_name_to_index_mapping.update(new_symbol)
    with open(
        f"{config.pytse_dir}/data/symbols_name.json", "r+", encoding="utf8"
    ) as file:
        data = json.load(file)
        data.update(new_symbol)
        file.seek(0)
        json.dump(data, file, ensure_ascii=False, indent=2)



from dataclasses import dataclass


@dataclass
class ShareholderData:
    id: str
    name: str
    instrument_id: str  # it's cIsin key in response
    shares: float
    percentage: float
    change: float



"""
This example is about downloading all financial indexes csv data.
"""

from pytse_client.download import download_financial_indexes

download_financial_indexes(symbols="all", write_to_csv=True, base_path="hello")



import json
from pathlib import Path
from typing import List
import locale

from pytse_client import config
from pytse_client.scraper.symbol_scraper import (
    MarketSymbol,
    get_market_symbols_from_market_watch_page,
    get_market_symbols_from_symbols_list_page,
    add_old_indexes_to_market_symbols,
)


def write_symbols_to_json(
    market_symbols: List[MarketSymbol], filename: str, path: str
) -> None:
    Path(path).mkdir(parents=True, exist_ok=True)
    with open(f'{path}/{filename}', 'w', encoding='utf8') as file:
        data = {
            obj.symbol: {
                "index": obj.index,
                "code": obj.code,
                "name": obj.name,
                "old": obj.old
            }
            for obj in market_symbols
        }
        json.dump(data, file, ensure_ascii=False, indent=2)


if __name__ == "__main__":
    # the sum order is important
    # https://github.com/Glyphack/pytse-client/issues/123
    market_symbols = (
        get_market_symbols_from_market_watch_page() +
        get_market_symbols_from_symbols_list_page()
    )
    deduplicated_market_symbols = list(set(market_symbols))
    # fetch old indexes of symbols
    deduplicated_market_symbols = add_old_indexes_to_market_symbols(
        deduplicated_market_symbols
    )
    # sort by sybmol
    locale.setlocale(locale.LC_COLLATE, "fa_IR.UTF-8")
    sorted_market_symbols = sorted(deduplicated_market_symbols)
    write_symbols_to_json(
        sorted_market_symbols, "symbols_name.json", f"{config.pytse_dir}/data"
    )



# utilities to extract data from APIs related to Ticker module

from collections import namedtuple
from dataclasses import dataclass
from typing import List, Tuple

Order = namedtuple('order', ['count', 'volume', 'price'])


@dataclass
class TradeSummary:
    buy_vol: float
    buy_count: float
    sell_vol: float
    sell_count: float


def get_orders(orders_raw_text: str) -> Tuple[List[Order], List[Order]]:
    if not orders_raw_text:
        return [], []
    buy_orders_set = []
    sell_orders_set = []
    orders = orders_raw_text.split(',')
    orders.pop()  # last item is empty string
    for order_text in orders:
        order_numbers = order_text.split("@")
        buy_orders_set.append(
            Order(
                order_numbers[0],  # count
                order_numbers[1],  # vol
                order_numbers[2],  # price
            )
        )
        sell_orders_set.append(
            Order(
                order_numbers[5],  # count
                order_numbers[4],  # vol
                order_numbers[3],  # price
            )
        )
    return buy_orders_set, sell_orders_set


def get_individual_trade_summary(
    individual_trade_summary_section
) -> TradeSummary:
    splitted_fields = individual_trade_summary_section.split(",")
    if len(splitted_fields) < 9:
        return None

    individual_buy_vol = float(splitted_fields[0])
    individual_sell_vol = float(splitted_fields[3])
    individual_buy_count = float(splitted_fields[5])
    individual_sell_count = float(splitted_fields[8])

    return TradeSummary(
        buy_vol=individual_buy_vol,
        buy_count=individual_buy_count,
        sell_vol=individual_sell_vol,
        sell_count=individual_sell_count,
    )


def get_corporate_trade_summary(corporate_trade_summary_section):
    splitted_fields = corporate_trade_summary_section.split(",")
    if len(splitted_fields) < 9:
        return None

    corporate_buy_vol = float(splitted_fields[1])
    corporate_sell_vol = float(splitted_fields[4])
    corporate_buy_count = float(splitted_fields[6])
    corporate_sell_count = float(splitted_fields[9])

    return TradeSummary(
        buy_vol=corporate_buy_vol,
        buy_count=corporate_buy_count,
        sell_vol=corporate_sell_vol,
        sell_count=corporate_sell_count,
    )



from .download import download
from .download import download_client_types_records
from .download import download_financial_indexes
from .key_stats import get_aggregated_key_stats
from .symbols_data import all_symbols
from .ticker import Ticker
from .financial_index import FinancialIndex



"""
This example is about downloading two sample financial indexes.
"""

from pytse_client.download import download_financial_indexes

download_financial_indexes(
    symbols=["شاخص قيمت 50 شركت", "فني مهندسي"],
    write_to_csv=True,
    include_jdate=True
)



import os

FINANCIAL_INDEX_BASE_PATH = "financial_index_data"
DATA_BASE_PATH = "tickers_data"
KEY_STATS_BASE_PATH = "key_stats_data"
CLIENT_TYPES_DATA_BASE_PATH = "client_types_data"
SHAREHOLDERS_HISTORY_PATH = "shareholders_data"
pytse_dir = os.path.dirname(__file__)
LOGGER_NAME = "pytse"



import asyncio
import time
from typing import List


async def run_tasks_with_wait(tasks: List, chunk_size: int, wait_between: int):
    """
    This function runs async tasks chunk by chunk and
    wait after completing each chunk
    """
    chunks = chunk_it(tasks, chunk_size)
    final_result = []
    for chunk in chunks:
        final_result.extend(await asyncio.gather(*chunk))
        time.sleep(wait_between)
    return final_result


def chunk_it(list_to_split, chunk_size):
    list_of_chunks = []
    start_chunk = 0
    end_chunk = start_chunk + chunk_size
    while end_chunk <= len(list_to_split) + chunk_size:
        chunk_ls = list_to_split[start_chunk:end_chunk]
        list_of_chunks.append(chunk_ls)
        start_chunk = start_chunk + chunk_size
        end_chunk = end_chunk + chunk_size
    return list_of_chunks



import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry


def requests_retry_session(
        retries=10,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504, 503),
        session=None,
) -> requests.Session:
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



from datetime import datetime, timedelta
from typing import Generator


def datetime_range(
    start=None, end=None
) -> Generator[datetime.date, None, None]:
    span = end - start
    for i in range(span.days + 1):
        yield start + timedelta(days=i)



import locale
import logging
import re
from dataclasses import dataclass
from typing import List

import requests
from bs4 import BeautifulSoup
from pytse_client import config, tse_settings
from pytse_client.utils import requests_retry_session
from pytse_client.utils.persian import replace_arabic
from requests import HTTPError

logger = logging.getLogger(config.LOGGER_NAME)


@dataclass
class MarketSymbol:
    code: str
    symbol: str
    name: str
    index: int
    old: List[int]

    def __hash__(self):
        """Hash function is used for deduplication"""
        return hash(self.symbol)

    def __lt__(self, other):
        return not locale.strcoll(other.symbol, self.symbol) < 0

    def __eq__(self, other):
        return locale.strcoll(other.symbol, self.symbol) == 0


def get_market_symbols_from_symbols_list_page() -> List[MarketSymbol]:
    """uses SYMBOLS_LIST_URL and scrapes symbols information from the page

    :return: list of all market symbols
    :rtype: List[MarketSymbol]
    """
    url = tse_settings.SYMBOLS_LIST_URL
    soup = BeautifulSoup(requests.get(url).content, 'html.parser')
    table = soup.find("table")
    market_symbols = []
    # Theres no thead or tbody in this table tag
    # so we eliminate first loop since its the table headers.
    table_rows = table.find_all('tr')[1:]
    for table_row in table_rows:
        row_data = table_row.find_all('td')
        # escape old symbols
        if row_data[7].a.text.startswith('حذف-'):
            continue
        market_symbols.append(
            MarketSymbol(
                code=row_data[0].text,
                symbol=replace_arabic(row_data[6].a.text),
                name=replace_arabic(row_data[7].a.text).replace('\u200c', ''),
                index=row_data[6].a.get('href').partition('inscode=')[2],
                old=[],
            )
        )
    return market_symbols


def get_market_symbols_from_market_watch_page() -> List[MarketSymbol]:
    """get symbols list from market watch page

    :return: list of market symbols
    :rtype: List[MarketSymbol]
    """
    response = requests.get(tse_settings.MARKET_WATCH_INIT_URL)

    # response contain different groups for different data
    response_groups = response.text.split("@")

    if len(response_groups) < 3:
        logger.error(
            "symbols information from market watch page is not valid",
            extra={"response": response}
        )

    symbols_data = response_groups[2].split(";")

    market_symbols = []
    for symbol_data in symbols_data:
        data = symbol_data.split(",")
        symbol_name_ends_with_number = re.search(r'\d+$', data[2])

        # if symbol name ends with number it's some kind of symbol
        # like 'حق تقدم' and we don't want it
        if symbol_name_ends_with_number:
            continue

        market_symbols.append(
            MarketSymbol(
                code=replace_arabic(data[1]),
                symbol=replace_arabic(data[2]),
                index=replace_arabic(data[0]),
                name=replace_arabic(data[3]).replace('\u200c', ''),
                old=[],
            )
        )

    return market_symbols


def add_old_indexes_to_market_symbols(
    symbols: List[MarketSymbol]
) -> List[MarketSymbol]:
    """
    get old symbols list from search page

    Parameters
    ----------
    market_symbols : List[MarketSymbol]
        list of market symbols.

    Returns
    -------
    List[MarketSymbol]
        list of market symbols.

    """
    market_symbols = []

    for symbol in symbols:
        index, old_ids = get_symbol_ids(symbol.symbol)
        if index is None:
            index = symbol.index
        market_symbols.append(
            MarketSymbol(
                code=symbol.code,
                symbol=symbol.symbol,
                index=index,
                name=symbol.name,
                old=old_ids,
            )
        )

    return market_symbols


def get_symbol_ids(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception(f"{symbol_name}: Sorry, tse server did not respond")

    symbols = response.text.split(';')
    index = None
    old_ids = []
    for symbol_full_info in symbols:
        if symbol_full_info.strip() == "":
            continue
        symbol_full_info = symbol_full_info.split(',')
        if replace_arabic(symbol_full_info[0]) == symbol_name:
            if symbol_full_info[7] == '1':
                index = symbol_full_info[2]  # active symbol id
            else:
                old_ids.append(symbol_full_info[2])  # old symbol id

    return index, old_ids



import pytse_client as tse

ticker = tse.Ticker("وبملت")
history = ticker.history


def sma(series, periods: int, ):
    return series.rolling(window=periods, min_periods=periods).mean()


sma_10 = sma(history.close, 10)
sma_20 = sma(history.close, 20)
buy_signals = (
        (sma_10 > sma_20) &
        (sma_20.shift(1) > sma_10.shift(1))
)
print(buy_signals.tail(60))



import functools


def catch(*exception_args):

    def wrapper(func):

        @functools.wraps(func)
        def inner(*args, **kwargs):

            try:
                return func(*args, **kwargs)

            except exception_args:
                return None

        return inner

    return wrapper



def get_index_to_symbol_map(dic: dict):
    """
    convert {<symbol>: {index:<index>,...}}
    to
    {<index>: {symbol:<symbol>,...}}
    """
    tmp_dic = {item[1]["index"]: {**item[1], "symbol": item[0]}
               for item in dic.items()}
    final_dic = {item[0]: {item2[0]: item2[1] for item2 in item[1].items(
    ) if item2[0] != "index"} for item in tmp_dic.items()}
    return final_dic


if __name__ == '__main__':
    d = {"s": {"index": 1, "dummy": 100}, "a": {
        "index": 2, "dummy": 100}, "j": {"index": 3, "dummy": 100}}
    print(get_index_to_symbol_map(d))
    # output >>
    # {
    # 1: {'dummy': 100, 'symbol': 's'},
    # 2: {'dummy': 100, 'symbol': 'a'},
    # 3: {'dummy': 100, 'symbol': 'j'}
    # }



from pytse_client.utils.request_session import requests_retry_session
from pytse_client.utils.scrape import (
    get_html_table_header_and_rows,
    get_shareholders_html_table_as_csv,
)
from pytse_client.utils.string import convert_to_number_if_number
from pytse_client.utils.date import datetime_range



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytse_client',
 'pytse_client.data',
 'pytse_client.examples',
 'pytse_client.financial_index',
 'pytse_client.proxy',
 'pytse_client.scraper',
 'pytse_client.scripts',
 'pytse_client.ticker',
 'pytse_client.ticker_statisticals',
 'pytse_client.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'jdatetime>=3.6.2,<4.0.0',
 'lxml>=4.6.3,<5.0.0',
 'pandas',
 'requests>=2.23.0,<3.0.0',
 'tenacity>=7.0.0,<8.0.0']

setup_kwargs = {
    'name': 'pytse-client',
    'version': '0.12.0',
    'description': 'tehran stock exchange(TSE) client in python',
    'long_description': '<div dir="rtl"  markdown="1">\n\n# دریافت اطلاعات بازار بورس تهران\n\n![Python application](https://github.com/Glyphack/pytse-client/workflows/Python%20application/badge.svg)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/Glyphack/pytse-client.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Glyphack/pytse-client/context:python)\n[![Discord Chat](https://img.shields.io/discord/730808323808559106?label=discord)](https://discord.gg/ampPDKHpVv)\n\nبا استفاده از pytse client میتونید به دیتای بازار بورس تهران در پایتون دسترسی داشته باشید.\nهدف حل مشکلات گرفتن اطلاعات بروز از سایت بازار بورس تهران هست.\n\n## میخواید مشارکت کنید؟\n\nلطفا [این صفحه](https://github.com/Glyphack/pytse-client/blob/master/CONTRIBUTING.md) رو مطالعه کنید\n\n- [دریافت اطلاعات بازار بورس تهران](#دریافت-اطلاعات-بازار-بورس-تهران)\n  - [میخواید مشارکت کنید؟](#میخواید-مشارکت-کنید)\n  - [قابلیت\u200cها](#قابلیتها)\n  - [نصب](#نصب)\n  - [نصب آخرین نسخه در حال توسعه](#نصب-آخرین-نسخه-در-حال-توسعه)\n  - [نحوه استفاده](#نحوه-استفاده)\n    - [دانلود سابقه سهم ها](#دانلود-سابقه-سهم-ها)\n    - [دانلود سابقه شاخص های مالی](#دانلود-سابقه-شاخص-های-مالی)\n    - [دانلود سابقه معاملات حقیقی و حقوقی به صورت مجزا](#دانلود-سابقه-معاملات-حقیقی-و-حقوقی-به-صورت-مجزا)\n    - [ماژول Ticker](#ماژول-ticker)\n      - [نکته ۱](#نکته-۱)\n      - [نکته ۲](#نکته-۲)\n      - [اطلاعات نماد\u200cهای حذف شده](#اطلاعات-نمادهای-حذف-شده)\n      - [اطلاعات حقیقی و حقوقی](#اطلاعات-حقیقی-و-حقوقی)\n      - [سهامداران عمده](#سهامداران-عمده)\n        - [شناور سهم](#شناور-سهم)\n        - [تاریخچه\u200cی سهام\u200cداران عمده](#تاریخچهی-سهامداران-عمده)\n        - [اطلاعات لحظه\u200cای سهام](#اطلاعات-لحظهای-سهام)\n        - [ریز معاملات سهام](#ریز-معاملات-سهام)\n    - [آمارهای کلیدی نمادها](#آمارهای-کلیدی-سهام)\n  - [کامیونیتی](#کامیونیتی)\n  - [پکیج\u200cهای مورد نیاز:](#پکیجهای-مورد-نیاز)\n  - [الهام گرفته از:](#الهام-گرفته-از)\n\n## قابلیت\u200cها\n\n- دریافت اطلاعات تاریخی به صورت تعدیل\u200cشده و تعدیل نشده برای نمادها\n- دریافت اطلاعات لحظه نمادها\n- دریافت اطلاعات تاریخی شاخص\u200cهای مالی\n- قابلیت گرفتن اطلاعات یک سهم مانند گروه سهم و اطلاعات معاملات حقیقی و حقوقی\n- دریافت اطلاعات فاندامنتال یک نماد شامل (EPS ,P/E و حجم مبنا)\n- دریافت اطلاعات سهامداران عمده\n- دریافت آمارهای کلیدی مربوط به فیلترنویسی برای نمادها\n- دریافت ریزمعاملاتی آخرین روز معاملاتی نمادها\n\n## نصب\n\n<div dir="ltr">\n\n```bash\npip install pytse-client\n```\n\n</div>\n\n## نصب آخرین نسخه در حال توسعه\n\nاین نسخه\u200cی در حال توسعه است که بر روی گیت\u200cهاب قرار دارد، همه\u200cی قابلیت\u200cهای گفته شده در این صفحه را دارد اما ممکن است بعضی قابلیت\u200cهای جدید تست شده نباشند. در صورتی که نسخه\u200cی بالا کاری که میخواهید را انجام نمیدهد این را نصب کنید.\n\n<div dir="ltr">\n\n```bash\npip install git+https://github.com/Glyphack/pytse-client.git\n```\n\n</div>\n\n## نحوه استفاده\n\n### دانلود سابقه سهم ها\n\nبا استفاده از این تابع می\u200cتوان سابقه سهام رو دریافت کرد و هم اون رو ذخیره و هم توی کد استفاده کرد\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\ntickers = tse.download(symbols="all", write_to_csv=True)\nprint(tickers["ولملت"])  # history\n\n# Output\n            date     open     high  ...     volume  count    close\n0     2009-02-18   1050.0   1050.0  ...  330851245    800   1050.0\n1     2009-02-21   1051.0   1076.0  ...  335334212   6457   1057.0\n2     2009-02-22   1065.0   1074.0  ...    8435464    603   1055.0\n3     2009-02-23   1066.0   1067.0  ...    8570222    937   1060.0\n4     2009-02-25   1061.0   1064.0  ...    7434309    616   1060.0\n...          ...      ...      ...  ...        ...    ...      ...\n2323  2020-04-14   9322.0   9551.0  ...  105551315  13536   9400.0\n2324  2020-04-15   9410.0   9815.0  ...  201457026  11322   9815.0\n2325  2020-04-18  10283.0  10283.0  ...  142377245   8929  10283.0\n2326  2020-04-19  10797.0  10797.0  ...  292985635  22208  10380.0\n2327  2020-04-20  10600.0  11268.0  ...  295590437  16313  11268.0\n```\n\n</div>\n\nبرای دانلود سابقه یک یا چند سهم کافیست اسم اون\u200cها به تابع داده بشه:\n\n<div dir="rtl">\n\nهمچنین با گذاشتن\n`write_to_csv=True`\nسابقه سهم توی فایلی با نماد سهم نوشته میشه\n\nسابقه سهم در قالب `Dataframe` است\n\nدر صورتی که می\u200cخواهید تاریخ شمسی به خروجی اضافه شود می\u200cتوانید با گذاشتن\n`include_jdate=True`\nاین امکان را فراهم کنید\n\nبرای دریافت قیمت\u200cهای تعدیل شده از\n`adjust=True`\nاستفاده کنید\n\n</div>\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\ntse.download(symbols="وبملت", write_to_csv=True)\ntse.download(symbols="وبملت", write_to_csv=True, include_jdate=True)\ntse.download(symbols=["وبملت", "ولملت"], write_to_csv=True)\n```\n\n</div>\n\n### دانلود سابقه شاخص های مالی\n\n<div style="line-height:80px;">\nبرای دانلود سابقه شاخص های منتخب بازار که از طریق این <a href="http://www.tsetmc.com/Loader.aspx?Partree=151315&Flow=1">  لینک </a>\nمی توانید لیست نام آن ها را ملاحظه کنید کافی است اسم شاخص یا شاخص های مورد نظر خود را به صورت\n<strong> دقیقا مشابه با لینک ارائه شده در بالا </strong>\n\nدر بخش \u200d\u200d\n`symbols`\nوارد کنید.\n\nهمینطور در صورتی که علاقه داشته باشید دیتای مربوط به همه شاخص ها را دریافت کنید کافی است که در برابر کلید `symbols`\nرشته\n`"all"`\nوارد کنید.\n\nچند نمونه از دریافت داده\u200cهای شاخصی را می\u200cتوانید در پایین مشاهده کنید.\n\n</div>\n<div dir="ltr">\n\n```python\nfrom pytse_client.download import download_financial_indexes\n\ndownload_financial_indexes(symbols="all", write_to_csv=True, base_path="hello")\n\n\ndownload_financial_indexes(symbols=["شاخص قيمت 50 شركت", "فني مهندسي"], write_to_csv=True, include_jdate=True)\n```\n\n</div>\n<div>\n<i>\nبه این موضوع توجه داشته باشید که دیتای دریافتی تنها حاوی مقدار شاخص مورد نظر در پایان روز معاملاتی است و اطلاعاتی از `OHLCV`\nشاخص در اختیار شما قرار نمی\u200cدهد.\n</i>\n</div>\n\n### دانلود سابقه معاملات حقیقی و حقوقی به صورت مجزا\n\nبرای دانلود سابقه معاملات حقیقی و حقوقی برای تمامی نمادها می\u200cتوان از تابع زیر استفاده کرد\n\n<div dir="ltr">\n\n```python\nfrom pytse_client import download_client_types_records\n\nif __name__ == \'__main__\':\n    records_dict = download_client_types_records("all")\n    print(records_dict["فولاد"])\n\n# Output\ndate         individual_buy_count  ... individual_ownership_change\n\n2020-09-01                36298  ...                   -691857.0\n2020-08-31                58185  ...                  83789408.0\n2020-08-26                  461  ...                  21647730.0\n2020-08-25                 1248  ...                  14716846.0\n2020-08-24                38291  ...                -238454702.0\n...                         ...  ...                         ...\n2008-12-02                    7  ...                    -10000.0\n2008-12-01                    8  ...                         0.0\n2008-11-30                   10  ...                    -12781.0\n2008-11-29                  116  ...                   4596856.0\n2008-11-26                   14  ...                    -20000.0\n\n[2518 rows x 17 columns]\n\n```\n\n</div>\n\nمشابه تابع قبلی می\u200cتوان نتایج را ذخیره کرد\n\n<div dir="ltr">\n\n```python\nfrom pytse_client import download_client_types_records\n\nif __name__ == \'__main__\':\n    # Records are saved as a .csv file with the same name of ticer\'s\n    records = download_client_types_records("فولاد", write_to_csv=True)\n```\n\n</div>\n\n### ماژول Ticker\n\nاین ماژول برای کار با دیتای یک سهم خاص هست و با گرفتن نماد اطلاعات موجود رو میده\n\nبرای مثال:\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\ntse.download(symbols="نوری", write_to_csv=True)  # optional\nticker = tse.Ticker("نوری")\n\nprint(ticker.history)  # سابقه قیمت سهم\nprint(ticker.client_types)  # حقیقی حقوقی\nprint(ticker.title)  # نام شرکت\nپتروشيمي نوري (نوري)\nprint(ticker.url)  # آدرس صفحه سهم\nhttp://tsetmc.com/Loader.aspx?ParTree=151311&i=19040514831923530\nprint(ticker.group_name)  # نام گروه\nمحصولات شيميايي\nprint(ticker.fiscal_year)  # سال مالی\n12/29\nprint(ticker.eps)  # EPS\n16442.0\nprint(ticker.p_e_ratio)  # P/E\n6.705388638851721\nprint(ticker.group_p_e_ratio)  # group P/E\n8.24\nprint(ticker.nav)  # NAV خالص ارزش دارایی\u200cها ویژه صندوق\u200cها می\u200cباشد\n112,881\nprint(ticker.nav_date)  # last date of NAV تاریخ بروزرسانی خالص ارزش دارایی\u200cها ویژه صندوق\u200cها می\u200cباشد\n1400/7/25 13:58:00\nprint(ticker.psr)  # PSR این نسبت ویژه شرکت\u200cهای تولیدی است\n71483.0238888889\nprint(ticker.p_s_ratio)  # P/S این نسبت ویژه شرکت\u200cهای تولیدی است\n1.5423242331125966\nprint(ticker.base_volume)  # حجم مبنا\n918780.0\nprint(ticker.state)  # وضعیت نماد\nمجاز\nprint(ticker.last_price)  # آخرین معامله\n109940\nprint(ticker.adj_close)  # قیمت پایانی\n110250\nprint(ticker.yesterday_price)  # قیمت دیروز\n106800\nprint(ticker.open_price)  # قیمت اولین معامله\n108200\nprint(ticker.high_price)  # قیمت حداکثر\n111830\nprint(ticker.low_price)  # قیمت حداقل\n108200\nprint(ticker.count)  # تعداد معاملات\n3934\nprint(ticker.volume)  # حجم معاملات\n2602437\nprint(ticker.value)  # ارزش معاملات\n286919407590\nprint(ticker.last_date)  # تاریخ آخرین اطلاعات قیمت پایانی ناشی از تغییرات شرکتی و معاملات\n2021-11-01 12:29:54\nprint(ticker.flow)  # عنوان بازار\nبورس\nprint(ticker.sta_max)  # حداکثر قیمت مجاز\n115760.0\nprint(ticker.sta_min)  # حداقل قیمت مجاز\n104740.0\nprint(ticker.min_week)  # حداقل قیمت هفته اخیر\n104500.0\nprint(ticker.max_week)  # حداکثر قیمت هفته اخیر\n111830.0\nprint(ticker.min_year)  # حداقل قیمت بازه سال\n48320.0\nprint(ticker.max_year)  # حداکثر قیمت بازه سال\n197000.0\nprint(ticker.month_average_volume)  # میانگین حجم ماه\n3865804\nprint(ticker.float_shares)  # درصد سهام شناور\n10.0\nprint(ticker.best_supply_price)  # قیمت بهترین تقاضا\n109960\nprint(ticker.best_supply_vol)  # حجم بهترین تقاضا\n8296\nprint(ticker.best_demand_price)  # قیمت بهترین عرضه\n109920\nprint(ticker.best_demand_vol)  # حجم بهترین عرضه\n3620\nprint(ticker.total_shares)  # تعداد سهام\nprint(ticker.market_cap) # ارزش بازار شرکت\nprint(ticker.shareholders)  # اطلاعات سهام\u200cداران عمده\nprint(ticker.get_shareholders_history())) # تاریخچه\u200cی سهام\u200cداران عمده\nprint(ticker.get_trade_details())  # ریز معاملات روز جاری\nprint(ticker.get_ticker_real_time_info_response()) # اطلاعات لحظه\u200cای مانند قیمت و پیشنهادات خرید و فروش\n```\n\n</div>\nبرای دریافت قیمت\u200cهای تعدیل شده هم میشه از این کد استفاده کرد\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker(symbol="بركت", adjust=True)\n```\n\n</div>\n\nبرای استفاده لازم نیست حتما تابع دانلود صدا زده بشه.\nاگر این کد رو بدون دانلود کردن سهم استفاده کنید خودش اطلاعات سهم رو از سایت میگیره،\nاما اگر قبل از اون از دانلود استفاده کرده باشید\nبه جای گرفتن از اینترنت اطلاعات رو از روی فایل میخونه که سریع تر هست\n\n##### نکته ۱\n\nطبق تجربه\u200c ای که داشتم چون گاهی اوقات سایت بورس مدت زیادی طول میکشه تا اطلاعات رو بفرسته یا بعضی مواقع نمی\u200cفرسته بهتر هست که اول تابع دانلود رو استفاده کنید برای سهم\u200cهایی که لازم هست و بعد با دیتای اون\u200cها کار کنید.\n\nدر صورت نیاز به اطلاعات لحظه\u200cای نماد بهتر است کل [اطلاعات لحظه\u200cای سهام](#اطلاعات-لحظهای-سهام) را یکجا دریافت کنید تا هم دیتای دریافتی مربوط به یک زمان باشند و هم از ارسال درخواست\u200cهای مکرر به سایت بورس جلوگیری شود.\n\n##### نکته ۲\n\nبعضی از ویژگی\u200cها برای همه\u200cی سهم\u200cها در دسترس نیست. برای مثال بعضی از سهم\u200cها دارای آخرین قیمت یا پی به ای یا ای پی اس نیستند. مقدار این ویژگی\u200cها در صورت نبودن برابر با `None` خواهد بود. پس باید در برنامه خود اینکه این مقادیر وجود دارند را بررسی کنید.\n\n#### اطلاعات نماد\u200cهای حذف شده\n\nتعدادی از نماد\u200cها توی سایت به شکل حذف شده هستند. برای گرفتن دیتای این نماد\u200cها از ماژول تیکر استفاده کنید.\nبرای مثال جهت دسترسی به دیتای نماد حذف شده خصدرا، اندیس آن را از آدرس نماد در سایت بورس بگیرید\nhttp://www.tsetmc.com/Loader.aspx?ParTree=151311&i=25165947991415904\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("", index="25165947991415904")\n```\n\n</div>\n\nمقدار \u200d`index` را با مقدار جلوی `i=` جایگزین میکنیم.\n\n#### اطلاعات حقیقی و حقوقی\n\nاطلاعات خرید و فروش حقیقی و حقوقی سهام رو میشه از طریق `ticker.client_types` گرفت این اطلاعات یه DataFrame شامل اطلاعات موجود در تب حقیقی حقوقی(تب بنفشی که در این [صفحه](http://www.tsetmc.com/Loader.aspx?ParTree=151311&i=778253364357513) هست) سهم هست:\n\n<div dir="ltr">\n\n```\ndate : تاریخ\nindividual_buy_count : تعداد معاملات خرید حقیقی\ncorporate_buy_count : تعداد معلاملات خرید حقوقی\nindividual_sell_count : تعداد معاملات فروش حقیقی\ncorporate_sell_count : تعداد معلاملات فروش حقوقی\nindividual_buy_vol : حجم خرید حقیقی\ncorporate_buy_vol : حجم خرید حقوقی\nindividual_sell_vol : حجم فروش حقیقی\ncorporate_sell_value : حجم فروش حقوقی\nindividual_buy_mean_price : قیمت میانگین خرید حقیقی\nindividual_sell_mean_price : قیمت میانگین فروش حقیقی\ncorporate_buy_mean_price : قیمت میانگین خرید حقوقی\ncorporate_sell_mean_price : قیمت میانگین فروش حقوقی\nindividual_ownership_change : تغییر مالکیت حقوقی به حقیقی\n```\n\n</div>\n\n#### سهامداران عمده\n\nسهام\u200cداران عمده اطلاعات داخل این [صفحه](http://tsetmc.com/Loader.aspx?Partree=15131T&c=IRO1BMLT0007) هست.\nاین اطلاعات رو میشه با `shareholders` گرفت که یک DataFrame هست.\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("وبملت")\nprint(ticker.shareholders)  # اطلاعات سهام\u200cداران عمده\n\n# Output\n change   percentage       share                                 shareholder\n0   دولت جمهوري اسلامي ايران                    23,114,768,760  11.160     0\n1   صندوق تامين آتيه كاركنان بانك ملت           13,353,035,330  6.440      0\n2   صندوق سرمايه گذاري واسطه گري مالي يكم       11,748,764,647  5.670      0\n3   شركت پتروشيمي فن آوران-سهامي عام-           9,253,327,080   4.460      0\n4   شركت گروه مالي ملت-سهام عام-                8,933,698,834   4.310      0\n5   صندوق سرمايه گذاري.ا.بازارگرداني ملت     8,395,500,914   4.050      0\n6   شركت سرمايه گذاري صباتامين-سهامي عام-       7,659,597,269   3.690      0\n7   شركت تعاوني معين آتيه خواهان                4,561,801,327   2.200      0\n8   شركت س اتهران س.خ-م ك م ف ع-                4,278,903,677   2.060      0\n9   شركت گروه توسعه مالي مهرآيندگان-سهامي عام-  4,161,561,525   2.000      0\n10  شركت س اخراسان رضوي س.خ-م ك م ف ع-          3,442,236,423   1.660      0\n11  شركت س افارس س.خ-م ك م ف ع-                 2,593,956,288   1.250      0\n12  شركت س اخوزستان س.خ-م ك م ف ع-              2,526,080,803   1.220      0\n13  شركت شيرين عسل-سهامي خاص-                   2,496,936,881   1.200      0\n14  شركت سرمايه گذاري ملي ايران-سهامي عام-      2,423,674,676   1.170      0\n15  شركت س ااصفهان س.خ-م ك م ف ع-               2,274,221,331   1.090      0\n```\n\n</div>\n\n##### شناور سهم\n\nبرای مثال میشه با استفاده از دیتای سهامداران عمده، شناوری سهم رو حساب کرد:\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("وبملت")\n\nprint(ticker.shareholders.percentage.sum())  # جمع درصد سهام\u200cداران عمده\n53.63\n\nprint(100 - ticker.shareholders.percentage.sum())  # درصد سهام شناور\n46.37\n```\n\n</div>\n\n##### تاریخچه\u200cی سهام\u200cداران عمده\n\nبا استفاده از تابع get_shareholders_history میشه تاریخچه اطلاعات سهام\u200cداران عمده رو گرفت:\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("وبملت")\nticker.get_shareholders_history(\n    from_when=datetime.timedelta(days=90),  # تعداد روز\u200cهای گذشته که مقدار پیشفرض ۹۰ روز است\n    to_when=datetime.datetime.now(),  # تا چه تاریخی اطلاعات گرفته شود که پیشفرض امروز است\n    only_trade_days=True,  # فقط روز\u200cهای معاملاتی که پیشفرض بله است\n)\n```\n\n</div>\nخروجی این تابع یک دیتا فریم حاوی دیتای زیر است:\n<div dir="ltr">\n\n```\n,date,shareholder_id,shareholder_shares,shareholder_percentage,shareholder_instrument_id,shareholder_name,change\n0,2021-08-30 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n1,2021-08-30 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n2,2021-08-30 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n3,2021-08-30 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n4,2021-08-30 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n5,2021-08-31 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n6,2021-08-31 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n7,2021-08-31 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n8,2021-08-31 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n9,2021-08-31 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n10,2021-09-01 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n11,2021-09-01 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n12,2021-09-01 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n13,2021-09-01 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n14,2021-09-01 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n\n```\n\n</div>\n\n<div id="qa" />\nگرفتن این دیتا کار زمان بری هست (با توجه به تعداد روزی که لازم دارید) و سریع کردن کار با کد به راحتی امکان پذیر نیست. سعی نکنید با همزمان اجرا کردن این تابع برای سهم\u200cهای مختلف روند رو سریع\u200cتر کنید چون سایت ip رو بلاک میکنه.\nاگر موقع اجرای کد پیغام زیر را به تعداد زیاد گرفتید (مثلا هر ثانیه این پیغام اومد) یعنی آیپی شما توسط سایت بورس بلاک شده و چند دقیقه صبر کنید و دوباره ادامه بدید.\n<div dir="ltr">\n\n```\nRetrying pytse_client.ticker.ticker.Ticker._get_ticker_daily_info_page_response in 1.3127419515957892 seconds as it raised ClientResponseError: 500, message=\'Internal Server Error\', url=URL(\'http://cdn.tsetmc.com/Loader.aspx?ParTree=15131P&i=56574323121551263&d=20210220\').\n```\n\n</div>\n\n##### اطلاعات لحظه\u200cای سهام\n\nاز طریق تابع `get_ticker_real_time_info_response` میشه اطلاعات لحظه\u200cای سهام رو گرفت.\nنمونه\u200cی استفاده\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("وبملت")\nreal_time_data = ticker.get_ticker_real_time_info_response()\n\nprint(real_time_data.buy_orders) # پیشنهادات خرید\nprint(real_time_data.sell_orders) # پیشنهادات فروش\nprint(real_time_data.best_supply_price)  # قیمت بهترین تقاضا\nprint(real_time_data.best_supply_vol)  # حجم بهترین تقاضا\nprint(real_time_data.best_demand_price)  # قیمت بهترین عرضه\nprint(real_time_data.best_demand_vol)  # حجم بهترین عرضه\nprint(real_time_data.state)  # وضعیت نماد\nprint(real_time_data.last_price)  # قیمت آخرین معامله\nprint(real_time_data.adj_close)  # قیمت پایانی\nprint(real_time_data.yesterday_price)  # قیمت دیروز\nprint(real_time_data.open_price)  # قیمت اولین معامله\nprint(real_time_data.high_price)  # قیمت حداکثر\nprint(real_time_data.low_price)  # قیمت حداقل\nprint(real_time_data.count)  # تعداد معاملات\nprint(real_time_data.volume)  # حجم معاملات\nprint(real_time_data.value)  # ارزش معاملات\nprint(real_time_data.last_date)  # آخرین اطلاعات قیمت ناشی از تغییرات شرکتی و معاملات\n\n# پیشنهادات فروش\nfor sell_order in real_time_data.sell_orders:\n    print(sell_order.volume, sell_order.count, sell_order.price)\n\n# پیشنهادات خرید\nfor buy_order in real_time_data.buy_orders:\n    print(buy_order.volume, buy_order.count, buy_order.price)\n\n# اطلاعات خرید و فروش حقیقی و حقوقی\nprint(real_time_data.individual_trade_summary.buy_count)\nprint(real_time_data.individual_trade_summary.buy_vol)\nprint(real_time_data.individual_trade_summary.sell_count)\nprint(real_time_data.individual_trade_summary.sell_vol)\nprint(real_time_data.corporate_trade_summary.buy_count)\nprint(real_time_data.corporate_trade_summary.buy_vol)\nprint(real_time_data.corporate_trade_summary.sell_count)\nprint(real_time_data.corporate_trade_summary.sell_vol)\n```\n\n</div>\n\n##### ریز معاملات سهام\n\nاز طریق تابع `get_trade_details` می\u200cتوان ریز معاملات آخرین روز معاملاتی سهام را گرفت:\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("نوری")\ntrade_details = ticker.get_trade_details()\nprint(trade_details)\n\n# Output\n          date  volume     price\n0     09:00:20   10000  111900.0\n1     09:00:20    4480  111900.0\n2     09:00:20    3171  111900.0\n3     09:00:20    1647  111900.0\n4     09:00:20    1101  111900.0\n       ...     ...       ...\n6478  12:29:57    1163  116000.0\n6479  12:29:57    2159  116000.0\n6480  12:29:57     795  116000.0\n6481  12:29:58     257  116000.0\n6482  12:29:59     601  116000.0\n```\n\n</div>\n\n\n### آمارهای کلیدی سهام\n\nدر بخش دیده\u200cبان بازار امکان فیلترنویسی به زبان جاوااسکریپت وجود دارد. یکی از اطلاعات مهمی که معامله\u200cگران معمولا با اسفاده از آن فیلترنویسی می\u200cکنند آمارهای کلیدی سهام است که شامل تمامی موارد ذکر شده در لینک زیر است.\n\n[آمارهای کلیدی](./pytse_client/ticker_statisticals/README.md)\n\nممکن است گاهی برخی از این اطلاعات موجود نباشند که باید در برنامه کاربردی از وجود آن برای سهام مورد نظر اطمینان پیدا کنید.\nالبته باید دقت داشت اگر برخی از این اطلاعات گاهی برای برخی نمادها موجود نبود در خود دیده\u200cبان بازار هم موجود نبوده است.\n\nدر حال حاضر امکان دریافت کل اطلاعات بروز شده و لحظه\u200cای آمارهای کلیدی با استفاده از تکه کد زیر موجود است.\n\n\n<div dir="ltr">\n\n```python\nfrom pytse_client import get_aggregated_key_stats\n\nkey_stats = get_aggregated_key_stats(base_path="hello", to_csv=True)\n\n# Output\n\n   ave_numof_buyer_last_12_month  ave_numof_buyer_last_3_month  ave_numof_corporation_buyer_last_12_month  ...  symbol                          name            index\n0                          610.0                         345.0                                        1.0  ...   وسپهر  سرمایه گذاری مالی سپهرصادرات  114312662654155\n1                          816.0                         381.0                                        1.0  ...    شصدف                صنعتی دوده فام  204092872752957\n2                           92.0                          76.0                                        0.0  ...     فسا                  پتروشیمی فسا  318005355896147\n3                          298.0                         246.0                                        2.0  ...   فرآور          فرآوریموادمعدنیایران  408934423224097\n4                          407.0                         332.0                                        1.0  ...   سبزوا              سیمان لار سبزوار  611986653700161\n\n```\n\nهمانطور که در مثال(ناقص به دلیل کمبود جا) بالا می بینید خروجی این دستور یک دیکشنری است که آمارهای کلیدی تمامی نمادهایی که در پکیج معتبر هستند و دیده\u200cبان در اختیار کاربران قرار می\u200cدهد را در خود دارد.\n\nبه صورت همزمان اطلاعات در \u200d\u200d`hello/key_stats.csv` ذخیره می\u200cشود.\nدر صورتی که نمی\u200cخواهید خروجی csv ساخته شود کافی است که `to_csv==False` قرار دهید.\nهمینطور در صورت خالی گذاشتن `base_path` به جای آن `key_stats_data/key_stats.csv` استفاده می\u200cشود.\n\n\n</div>\n\n\n## کامیونیتی\n\nاگر درباره پکیج یا استفاده از اون سوالی دارید میتونید توی سرور دیسکورد بپرسید.\n\nhttps://discord.gg/ampPDKHpVv\n\n<div id="required-packages" />\n\n## پکیج\u200cهای مورد نیاز:\n\n- [Pandas](https://github.com/pydata/pandas)\n- [Requests](http://docs.python-requests.org/en/master/)\n- [jdatetime](https://github.com/slashmili/python-jalali)\n- [beautifulsoup4](https://pypi.org/project/beautifulsoup4)\n\n<div id="credits" />\n\n## الهام گرفته از:\n\n- [tehran_stocks](https://github.com/ghodsizadeh/tehran-stocks)\n- [yfinance](https://github.com/ranaroussi/yfinance)\n\n</div>\n',
    'author': 'glyphack',
    'author_email': 'sh.hooshyari@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



import asyncio
import datetime
import logging
from time import sleep

import aiohttp
from pytse_client import Ticker

# set logging level to info to see progress of getting data
logging.basicConfig(level=logging.INFO)


async def create_session(limit=3):
    """
    optional can change default session
    """
    conn = aiohttp.TCPConnector(limit=limit)
    session = aiohttp.ClientSession(connector=conn)
    return session


async def get_shareholders_history():
    symbols = [
        "شکبیر",
        "برکت",
    ]
    from_times = [
        datetime.timedelta(days=90),
        datetime.timedelta(days=150),
    ]
    for symbol in symbols:
        # limit is the number of parallel requests
        # can increase and decrease speed
        session = await create_session(limit=2)
        print(f"downloading {symbol}")
        shareholders_history = await Ticker(symbol
                                            ).get_shareholders_history_async(
                                                session=session
                                            )
        shareholders_history.to_csv(f"{str(symbol)}.csv")
        print(f"downloaded {symbol} complete")

        # sleep between to avoid getting blocked
        sleep(50)


asyncio.run(get_shareholders_history())



# میانگین: ave
# ارزش: val
# معاملات: trans
# رتبه: rank
# حجم: vol
# دفعات: numof
# روزانه: daily
# وزن: w
# قیمت: price
# منفی: neg
# مثبت: pos
# حقوقی: corporation
# حقیقی: individual
# بسته: close
# باز: open

class KeyStats():
    # name : id
    ave_val_trans_last_3_month = 1
    ave_val_trans_last_12_month = 2
    rank_val_trans_last_3_month = 3
    rank_val_trans_last_12_month = 4

    ave_vol_trans_last_3_month = 5
    ave_vol_trans_last_12_month = 6
    rank_vol_trans_last_3_month = 7
    rank_vol_trans_last_12_month = 8

    ave_numof_trans_last_3_month = 9
    ave_numof_trans_last_12_month = 10
    rank_numof_trans_last_3_month = 11
    rank_numof_trans_last_12_month = 12

    w_ave_price_last_day_normal = 13
    w_ave_price_last_day_base_vol = 14

    val_trans_last_day = 15
    vol_trans_last_day = 16
    numof_trans_last_day = 17

    numof_neg_days_last_3_month = 18
    numof_neg_days_last_12_month = 19
    percent_neg_days_last_3_month = 20
    percent_neg_days_last_12_month = 21
    rank_neg_days_last_3_month = 22
    rank_neg_days_last_12_month = 23

    numof_notrade_days_last_3_month = 24
    numof_notrade_days_last_12_month = 25

    numof_pos_days_last_3_month = 26
    numof_pos_days_last_12_month = 27
    percent_pos_days_last_3_month = 28
    percent_pos_days_last_12_month = 29
    rank_pos_days_last_3_month = 30
    rank_pos_days_last_12_month = 31

    numof_trade_days_last_3_month = 32
    numof_trade_days_last_12_month = 33

    rank_trade_days_last_3_month = 34
    rank_trade_days_last_12_month = 35

    val_company_last_day = 36
    rank_val_company_last_day = 37

    numof_open_days_last_3_month = 38
    numof_open_days_last_12_month = 39
    percent_open_days_last_3_month = 40
    percent_open_days_last_12_month = 41
    rank_open_days_last_3_month = 42
    rank_open_days_last_12_month = 43

    numof_close_days_last_3_month = 44
    numof_close_days_last_12_month = 45
    percent_close_days_last_3_month = 46
    percent_close_days_last_12_month = 47
    rank_close_days_last_3_month = 48
    rank_close_days_last_12_month = 49

    ave_vol_individual_buy_last_3_month = 50
    ave_vol_individual_buy_last_12_month = 51
    rank_vol_individual_buy_last_3_month = 52
    rank_vol_individual_buy_last_12_month = 53

    ave_vol_corporation_buy_last_3_month = 54
    ave_vol_corporation_buy_last_12_month = 55
    rank_vol_corporation_buy_last_3_month = 56
    rank_vol_corporation_buy_last_12_month = 57

    ave_numof_individual_buyer_last_3_month = 58
    ave_numof_individual_buyer_last_12_month = 59
    rank_numof_individual_buyer_last_3_month = 60
    rank_numof_individual_buyer_last_12_month = 61

    ave_numof_corporation_buyer_last_3_month = 62
    ave_numof_corporation_buyer_last_12_month = 63
    rank_numof_corporation_buyer_last_3_month = 64
    rank_numof_corporation_buyer_last_12_month = 65

    ave_numof_buyer_last_3_month = 66
    ave_numof_buyer_last_12_month = 67
    rank_numof_buyer_last_3_month = 68
    rank_numof_buyer_last_12_month = 69

    ave_vol_individual_sell_last_3_month = 70
    ave_vol_individual_sell_last_12_month = 71
    rank_vol_individual_sell_last_3_month = 72
    rank_vol_individual_sell_last_12_month = 73

    ave_vol_corporation_sell_last_3_month = 74
    ave_vol_corporation_sell_last_12_month = 75
    rank_vol_corporation_sell_last_3_month = 76
    rank_vol_corporation_sell_last_12_month = 77

    ave_numof_individual_seller_last_3_month = 78
    ave_numof_individual_seller_last_12_month = 79
    rank_numof_individual_seller_last_3_month = 80
    rank_numof_individual_seller_last_12_month = 81

    ave_numof_corporation_seller_last_3_month = 82
    ave_numof_corporation_seller_last_12_month = 83
    rank_numof_corporation_seller_last_3_month = 84
    rank_numof_corporation_seller_last_12_month = 85

    ave_numof_seller_last_3_month = 86
    ave_numof_seller_last_12_month = 87
    rank_numof_seller_last_3_month = 88
    rank_numof_seller_last_12_month = 89


filter_class = KeyStats()

# collect all var names of filter_class
keys = [attr for attr in dir(filter_class) if not attr.startswith("__")]
# convert filter_class to dict
filter_key_value = {getattr(filter_class, key): key for key in keys}
filter_value_NONE = {key: None for key in filter_key_value.values()}

if __name__ == '__main__':
    print(filter_key_value)



import datetime
import logging
from typing import Union

import aiohttp
import requests
from pytse_client import config, tse_settings
from pytse_client.proxy.dto import ShareholderData
from tenacity import retry, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)


@retry(
    wait=wait_random(min=3, max=5),
    before_sleep=before_sleep_log(logger, logging.ERROR)
)
async def get_day_shareholders_history(
    ticker_index: str,
    date: datetime.date,
    session: Union[requests.Session, aiohttp.ClientSession],
):
    formatted_date = date.strftime(tse_settings.DATE_FORMAT)
    async with session.get(
        tse_settings.SYMBOL_DAY_INFO_SHAREHOLDERS_DATA.format(
            index=ticker_index, date=formatted_date
        ),
    ) as response:
        response.raise_for_status()
        response_json = await response.json()
        logger.info(
            f"""fetched shareholders data for
             {ticker_index} date: {formatted_date}"""
        )
        result = []
        for shareholders_raw_data in response_json["shareShareholder"]:
            # api returns data for today and tomorrow so we skip tomorrow
            # sometimes when we request data for a day it returns it with
            # tomorrow date instead(bug) so we only check that the
            # date is greater than requested date
            if shareholders_raw_data["dEven"] > int(formatted_date):
                continue
            result.append(
                ShareholderData(
                    id=shareholders_raw_data["shareHolderID"],
                    name=shareholders_raw_data["shareHolderName"],
                    instrument_id=shareholders_raw_data["cIsin"],
                    shares=shareholders_raw_data["numberOfShares"],
                    percentage=shareholders_raw_data["perOfShares"],
                    change=shareholders_raw_data["change"],
                )
            )
        return result



TSE_TICKER_EXPORT_DATA_ADDRESS = (
    "http://tsetmc.com/tsev2/data/Export-txt.aspx?t=i&a=1&b=0&i={}"
)

TSE_FINANCIAL_INDEX_EXPORT_DATA_ADDRESS = (
    "http://www.tsetmc.com/tsev2/chart/data/Index.aspx?i={}&t=value"
)

TSE_FINANCIAL_INDEX_EXPORT_INTRADAY_DATA_ADDRESS = (
    "http://www.tsetmc.com/Loader.aspx?ParTree=15131J&i={}"
)


TSE_TICKER_ADDRESS = ("http://tsetmc.com/Loader.aspx?ParTree=151311&i={}")

# FIXME: c should be cSecVal (group code of instrument)
# remove e parameter when instrument has not NAV value
TSE_ISNT_INFO_URL = (
    "http://www.tsetmc.com/tsev2/data/instinfofast.aspx?i={}&c=0&e=1"
)
TSE_TICKER_INTRODUCTION_URL = (
    "http://www.tsetmc.com/Loader.aspx?Partree=15131V&s={}"
)
TSE_CLIENT_TYPE_DATA_URL = (
    "http://www.tsetmc.com/tsev2/data/clienttype.aspx?i={}"
)
TSE_SYMBOL_ID_URL = ("http://www.tsetmc.com/tsev2/data/search.aspx?skey={}")

TSE_TRADE_DETAIL_URL = (
    "http://www.tsetmc.com/tsev2/data/TradeDetail.aspx?i={}"
)

# use instrument id as argument
TSE_SHAREHOLDERS_URL = (
    "http://www.tsetmc.com/Loader.aspx?Partree=15131T&c={}"
)

# returns all the instrument data for a specific date
# date should be formatted like YYYYMMDD
INSTRUMENT_DAY_INFO_URL = (
    "http://cdn.tsetmc.com/Loader.aspx?ParTree=15131P&i={index}&d={date}"
)

# returns list of all symbols
SYMBOLS_LIST_URL = ("http://www.tsetmc.com/Loader.aspx?ParTree=111C1417")

# market watch init, has initial data for market watch like symbols name etc.
MARKET_WATCH_INIT_URL = (
    "http://www.tsetmc.com/tsev2/data/MarketWatchInit.aspx?h=0&r=0"
)

# get day share holder data
SYMBOL_DAY_INFO_SHAREHOLDERS_DATA = (
    "http://cdn.tsetmc.com/api/Shareholder/{index}/{date}"
)

# get statistical information for all tickers
KEY_STATS_URL = (
    "http://www.tsetmc.com/tsev2/data/InstValue.aspx?t=a"
)


DATE_FORMAT = "%Y%m%d"
MIN_DATE = 20010321



HISTORY_FIELD_MAPPINGS = {
    "<DTYYYYMMDD>": "date",
    "<FIRST>": "open",
    "<HIGH>": "high",
    "<LOW>": "low",
    "<LAST>": "close",
    "<VOL>": "volume",
    "<CLOSE>": "adjClose",
    "<OPENINT>": "count",
    "<VALUE>": "value",
    "<OPEN>": "yesterday",
}

SHAREHOLDERS_FIELD_MAPPINGS = {
    "سهامدار/دارنده": "shareholder",
    "سهم": "shares",
    "درصد": "percentage",
    "تغییر": "change",
}
